@@ -15,7 +15,7 @@ import Control.Exception.Annotated
 import qualified Control.Exception.Safe as Safe
 import Data.Annotation
 import GHC.Stack
-
+import Data.Typeable
 import Data.AnnotationSpec ()
 import Data.Maybe
 
@@ -28,7 +28,8 @@ data TestException = TestException
     deriving (Eq, Show, Exception)
 
 instance Eq SomeException where
-    e0 == e1 = show e0 == show e1
+    SomeException (e0 :: e0) == SomeException (e1 :: e1) =
+        typeOf e0 == typeOf e1 && show e0 == show e1
 
 pass :: Expectation
 pass = pure ()
@@ -38,6 +39,22 @@ emptyAnnotation = pure
 
 spec :: Spec
 spec = do
+    describe "toException" $ do
+        it "wraps inner in SomeException" $ do
+            toException (AnnotatedException [] TestException)
+                `shouldBe` do
+                    SomeException
+                        (AnnotatedException [] (SomeException TestException))
+        it "flattens annotations" $ do
+            let
+                exn =
+                    AnnotatedException ["hello"] $
+                        AnnotatedException ["goodbye"] TestException
+            toException exn
+                `shouldBe` do
+                    SomeException $
+                        AnnotatedException ["hello", "goodbye"] (SomeException TestException)
+
     describe "AnnotatedException can fromException a" $ do
         it "different type" $ do
             fromException (toException TestException)
@@ -105,6 +122,40 @@ spec = do
             action
                 `shouldThrow`
                     (userError "uh oh" ==)
+        it "includes a callstack location" $ do
+            let
+                action =
+                    throw TestException
+                        `catch`
+                            \TestException ->
+                                throw TestException
+            action
+                `Safe.catch`
+                        \(e :: AnnotatedException TestException) -> do
+                            annotations e
+                                `callStackFunctionNamesShouldBe`
+                                    [ "throw"
+                                    , "throw"
+                                    , "catch"
+                                    ]
+    describe "catches" $ do
+        it "has a callstack entry" $ do
+            let
+                action =
+                    throw TestException
+                        `catches`
+                            [ Handler $ \TestException ->
+                                throw TestException
+                            ]
+            action
+                `Safe.catch`
+                        \(e :: AnnotatedException TestException) -> do
+                            annotations e
+                                `callStackFunctionNamesShouldBe`
+                                    [ "throw"
+                                    , "throw"
+                                    , "catches"
+                                    ]
 
     describe "tryAnnotated" $ do
         let subject :: (Exception e, Exception e') => e -> IO (AnnotatedException e')
@@ -153,7 +204,7 @@ spec = do
                     `Safe.catch` \(e :: AnnotatedException TestException) -> do
                         annotations e
                             `callStackFunctionNamesShouldBe`
-                                ["throwWithCallStack"
+                                [ "throwWithCallStack"
                                 , "checkpointCallStack"
                                 ]
 
